// GWave.cpp: implementation of the GWave class.
//
//////////////////////////////////////////////////////////////////////
#include "GWave.h"

enum { WTOP=1, WLEFT=2, WRIGHT=4, WBOTTOM=8, 
	   WTOP_LEFT=3, WTOP_RIGHT=5, WBOTTOM_LEFT=10, WBOTTOM_RIGHT=12 };

static const int flgs[]={ WTOP, WLEFT, WRIGHT, WBOTTOM, 
	WTOP_LEFT, WTOP_RIGHT, WBOTTOM_LEFT, WBOTTOM_RIGHT };


GWave::GWave( int w, int h )
{
	nWidth=w;
	nHeight=h;

	pWorld=new WBYTE[ nWidth*nHeight ];
	//             	
	ppWave   =new LPWBYTE [ 2*(nWidth+nHeight) ];
	ppWaveTmp=new LPWBYTE [ 2*(nWidth+nHeight) ];


	WPOINT dxy2[]={ {0,-1},{-1,0},{1,0},{0,1}, {-1,-1},{1,-1},{-1,1},{1,1} };
	int   ofs2[]={ -nWidth, -1, 1, nWidth, -nWidth-1, -nWidth+1, nWidth-1, nWidth+1 };

	memcpy( dxy, dxy2, sizeof(dxy) );
	memcpy( ofs, ofs2, sizeof(ofs) );
}

//   
void GWave::DrawBorder()
{
	//  
	for( int i=1; i<nHeight-1; i++ )  {
		pWorld[i*nWidth]=NOPASS_POINT;
		pWorld[(i+1)*nWidth-1]=NOPASS_POINT;
	}
	//  
	for (int j=0; j<nWidth; j++) {
		pWorld[j]=NOPASS_POINT;
		pWorld[nWidth*(nHeight-1)+j]=NOPASS_POINT;
	}
}

//        pWorld:
// 1)      
// 2)       ,
//	   - DrawBorder().
//  - =  
// -1,  >nMaxCount
//  0,    

int GWave::GetWay( WPOINT *pDPoints, WBYTE nMaxCount, WPOINT ptStart, WPOINT ptEnd )
{
	pWorld[ptStart.y*nWidth+ptStart.x]=START_POINT;
	pWorld[ptEnd.y*nWidth+ptEnd.x]=END_POINT;
	ppWave[0]=&pWorld[ptEnd.y*nWidth+ptEnd.x];
	nWavePoints=1;

//  
	for( nWave=1; nWave<=nMaxCount; nWave++ )
	{
		if( GenerateWave() ) break;
		if( nWavePoints==0 ) return 0;
	}
	if( nWave>nMaxCount ) return -1;


//  	
	WBYTE *p=&pWorld[ ptStart.y*nWidth+ptStart.x];
	for( int i=0; i<nWave; i++ )
	{
		int minK, flg=0;
		WBYTE *minp=p;

		//   
		for(int k=0; k<4; k++ )
		{
			WBYTE ch=p[ofs[k]];
			if( ch != NOPASS_POINT )
			{
				flg|=flgs[k];
				if( *minp>ch ) {
					minp=p+ofs[k];
					minK=k;
				}
			}
		}		
		//   
		for( ; k<8; k++ )
		{
			WBYTE ch=p[ofs[k]];
			if( flg&flgs[k] && *minp>ch ) {
				minp=p+ofs[k];
				minK=k;
			}
		}
		
		p=minp;
		pDPoints->x=dxy[minK].x;
		pDPoints->y=dxy[minK].y;
		pDPoints++;
	}
	
	return nWave;
}



bool GWave::GenerateWave( )
{
	WBYTE *p;
	WBYTE **pp;
	int fDiag;
	bool bFound=false;

	pp=ppWaveTmp;

	//  nWave 
	//       

	for( int i=0; i<nWavePoints; i++ )
	{
		p=ppWave[i]-nWidth;

		//   	
		// 5 1 6
		// 2   3
		// 7 4 8

		//        
		//      

		fDiag=0;
		WBYTE ch;
		
		// TOP 
		ch=*p;
		if( ch==PASS_POINT ) { *p=nWave; *pp++=p; fDiag|=WTOP; } else 
			if( ch<=nWave ) fDiag|=WTOP; else
				if( ch==START_POINT ) { bFound=true; break; }
		p+=nWidth-1;
	
		// LEFT 
		ch=*p;
		if( ch==PASS_POINT ) { *p=nWave; *pp++=p; fDiag|=WLEFT; } else 
			if( ch<=nWave ) fDiag|=WLEFT; else
				if( ch==START_POINT ) { bFound=true; break; }
		p+=2;	

		// RIGHT
		ch=*p;
		if( ch==PASS_POINT ) { *p=nWave; *pp++=p; fDiag|=WRIGHT; } else 
			if( ch<=nWave ) fDiag|=WRIGHT; else
				if( ch==START_POINT ) { bFound=true; break; }
		p+=nWidth-1;	

		// BOTTOM
		ch=*p;
		if( ch==PASS_POINT ) { *p=nWave; *pp++=p; fDiag|=WBOTTOM; } else 
			if( ch<=nWave ) fDiag|=WBOTTOM; else
				if( ch==START_POINT ) { bFound=true; break; }
		p-=2*nWidth+1;

		
		//          
		//      ,     

		// TOP_LEFT
		if( fDiag&WTOP_LEFT )
			if( *p==PASS_POINT ) { *p=nWave; *pp++=p; } else 
				if( *p==START_POINT ) { bFound=true; break; }
		p+=2;

		// TOP_RIGHT
		if( fDiag&WTOP_RIGHT )
			if( *p==PASS_POINT ) { *p=nWave; *pp++=p; } else 
				if( *p==START_POINT ) { bFound=true; break; }
		p+=2*nWidth-2;	

		// BOTTOM_LEFT
		if( fDiag&WBOTTOM_LEFT )
			if( *p==PASS_POINT ) { *p=nWave; *pp++=p; } else 
				if( *p==START_POINT ) { bFound=true; break; }
		p+=2;

		// BOTTOM_RIGHT
		if( fDiag&WBOTTOM_RIGHT ) 
			if( *p==PASS_POINT ) { *p=nWave; *pp++=p; } else 
				if( *p==START_POINT ) { bFound=true; break; }
	}
	
	nWavePoints=pp-ppWaveTmp;
	pp=ppWave;
	ppWave=ppWaveTmp;
	ppWaveTmp=pp;

	//  nWave ,       
	//  ppWave,    nWavePoints 

	return bFound;
}



GWave::~GWave()
{
	delete pWorld;
	delete ppWave;
	delete ppWaveTmp;
}
